# Relative_RGB_color_mask_maker Ver. 0.53
# Ver. 0.53 (2023.11)
# Relative_RGB_color_mask_maker 8 & 16bit Version
# add tanceparency factor
# add luminosity mask (0.12)
# add Monochrome film support (0.13)
# add Mask Inverting Option for Exclude Selected Area (0.15)
# add function of making combined mask (0.20)
# bugfix (0.22)
# User Interface Renewal (0.25)
# add Luminosity mask preview (0.25c)
# add imp2 mixing ratio parameter (0.14) for preparing next version
# improve Luminosity mask preview (0.25d)
# in Mg/G mask, add "compare B only" option (0.26)
# channel level adjustment option (0.26a)
# add final image preview function (0.27)
# modifying preview windows construction for Linux (0.27a)
# add final image preview function with 2nd excluding mask (0.28)
# add Skew of Color sensitivitiy (0.40)
# bug fix for color / mono judgement (0.40)
# add contrast adjustment (0.45)
# applicable for 32-bit image(0.50)
# Bug fix for luminosity mask & 32bit image (0.51)
# 32 bit image & UI renewed (0.52)
# Renewal of UI (0.53)
# (c)2021-23 Ohnishi, Yasuo

from ij import IJ, ImagePlus
from ij import WindowManager as WM
from ij.gui import GenericDialog
from ij.io import OpenDialog
from ij import plugin
from ij.plugin import ChannelSplitter, Duplicator
from ij.plugin.filter import ThresholdToSelection
from ij.process import ImageConverter, ImageProcessor, ByteProcessor, \
    ShortProcessor, FloatProcessor
from loci.plugins import BF
from java.awt.event import AdjustmentListener, ItemListener, TextListener
import os
from org.python.core import codecs
codecs.setDefaultEncoding('utf-8')

#-----------------
def set_Lmax(imp):
    # set Maximum value of Lightness
    L_max = imp.getDisplayRangeMax()
    if int(L_max) > 65536.01:
        L_max = 16777215
    elif int(L_max) > 256.01:
        L_max = 65535
    elif int(L_max) > 1.01:
        L_max = 255
    else:
        L_max = 1.0
    return L_max
#-----------------
def bit32to16(imp, L_max):
    imp.getProcessor().multiply(65535 / L_max)
    imp.setDisplayRange(0, 65535)
    ImageConverter(imp).convertToGray16()
    return imp
#-----------------
def clipUnder0(imp, Lmax = 1.0):
    # Only in 32bit image, this process is applicable. 
    ip = imp.getProcessor()
    imp.getProcessor().setThreshold(-1.0 * Lmax, \
            0.0, ImageProcessor.NO_LUT_UPDATE)
    roi = ThresholdToSelection.run(imp)
    ip.setValue(0.0)
    ip.setRoi(roi)
    ip.fill(roi)
    imp.setProcessor(ip)
    imp.setDisplayRange(0.0, Lmax)
    return imp
#-----------------
def convertToLuminosity(impOrg):
    # convert RGB 3 color image to mono Luminosity image
    imp = Duplicator().run(impOrg)
    impChs = ChannelSplitter.split(imp)
    ic = plugin.ImageCalculator()
    impChs[0].getProcessor().multiply(0.2126)
    impChs[1].getProcessor().multiply(0.7151)
    impChs[2].getProcessor().multiply(0.0722)
    for i in impChs:
        i.getProcessor().setLut(None)
    impTmp1 = ic.run("Add created", impChs[0], impChs[1])
    impLuminosity = ic.run("Add created", impTmp1, impChs[2])
    impTmp1.close()
    imp.close()
    return impLuminosity
#--------------------------------------------
def grayscale_mask_maker(imp1, imp2, imp3, YTh, Mode, \
        blInvert, Show, colorCh, TrFac, imp2MixRatio):
    #if Mode == "32-bit":
    Lmax = set_Lmax(imp1)
    imp1.setDisplayRange(0.0, Lmax)

# this is making relative color mask routine
    """
    parameters
        imp1: Target Channel (B), imp2: Comparative Channel (G)
        imp3: Comparative Channnel 2 (R)
        YTh: Target Color Ch Threshold (originally Yellow Threshold)
        blInvert: Color invert Yes or No (originally invert for blue mask)
        TrFac: Transparency Factor
        imp2MixRatio: imp2 Mixing Ratio (to imp3) 0.0 to 1.0 Normally
                                                                fixed to 0.50
        Rel2B: compare to only B ch. True/False
    """
# Comparative (G) ch.1 x 0.5
    imp2_05 = Duplicator().run(imp2)
    Vstr = str(imp2MixRatio)  # imp2 mixing ratio
    IJ.run(imp2_05, "Multiply...", "value=" + Vstr)
    imp2_05.setDisplayRange(0.0, Lmax * imp2MixRatio)
    if Show == "Yes":
        imp2_05.show()
# Comparative (R) ch.2 ch. x 0.5
    imp3_05 = Duplicator().run(imp3) # imp3 mixing ratio
    Vstr = str(1.0 - imp2MixRatio)
    IJ.run(imp3_05, "Multiply...", "value=" + Vstr)
    imp3_05.setDisplayRange(0.0, Lmax * (1.0 - imp2MixRatio))
    if Show == "Yes":
        imp3_05.show()
# Comparative (G) Ch. 1 x 0.5 + (R) Ch.2 x 0.5
    ic = plugin.ImageCalculator()
    imp32 = ic.run("Add created", imp3_05, imp2_05)
    imp32.setDisplayRange(0.0, Lmax)

# Close Windows
    if Show == "Yes":
        imp3.hide()
        imp2.hide()
        imp1.hide()
        imp32.show()
        imp2_05.hide()
        imp3_05.hide()
    imp2_05.close()
    imp3_05.close()
# (Ch1(R) x 0.5 + Ch2(G) x 0.5) - Target Ch (B)
    ic = plugin.ImageCalculator()
# Make Target Color (Yellow) Image Files
# in each Threshold, Make 3 kinds of Image Files: x1, x2, x4
    V1 = YTh
# Target (Yellow) Mask Threshold
    if blInvert == True: # for Blue mask
        YTh = YTh * -1 # Invert the threshold value for Blue mask
    V2 = str(YTh / 255.0 * Lmax) # actual Yellow mask Threshold
        # for making gray scale mask
            # Set the Threshold in R + G Image
    IJ.run(imp32, "Add...","value=" + V2)
    V = str(abs(V1))
    if V1 >= 0:
        V= "+" + V
    else:
        V= "-" + V
    if blInvert == False: # for making Yellow mask
        impGraymask = ic.run("Subtract created", imp32, imp1)
                                            # Make Yellow Mask
        if colorCh == "Yellow/Blue":
            V = "Y" + V
        elif colorCh == "Cyan/Red":
            V = "Cy" + V
        elif colorCh == "Magenta/Green":
            V = "Mg" + V
    else:                 # for making Primary Color (Blue) mask
        impGraymask = ic.run("Subtract created", imp1, imp32)
                                            # Make Blue Mask
        if colorCh == "Yellow/Blue":
            V = "B" + V
        elif colorCh == "Cyan/Red":
            V = "R" + V
        elif colorCh == "Magenta/Green":
            V = "G" + V

    if Mode == "32-bit":
        impGraymask.setDisplayRange(0.0, Lmax)
    if Show == "Yes":
        impGraymask.show()  # Show the subtracted image in window
    if TrFac < 100: # making grayscale mask
        IJ.run(impGraymask, "Multiply...", "value=" + str(4 * TrFac))
            # making mask more transparent (x 4)
    else:
        # if TrFac = 100, making relative color mask binary (calling subroutine)
        impGtoBmask = make_binary_mask(impGraymask, 1, \
                            Lmax, Mode, Show)
        impGtoBmask.setDisplayRange(0.0, Lmax)

        if Show == "Yes":
            impGraymask.hide()  # Hide the subtracted image in window
        impGraymask = impGtoBmask
    imp32.hide()
    imp32.close()
    return impGraymask, V
    """
    return values
        impGraymask: mask image
        V: description for mask parameters
    """
#-------------------------------------------
def luminosity_mask_maker(imp, LTh, Mode, NegaInvert, Show, TrFac):
    imp1 = Duplicator().run(imp)
    Lmax = set_Lmax(imp)
    """
    parameters
        imp1: Grayscale image > Liminosity Mask image
        LTh: Zone lower Threshold
        NegaInvert: Nega invert True or False
        TrFac: Transparency Factor
    """
    if Show == "Yes":
        imp1.show()
# Make Grayscale Mask Image Files
    V1 = str(LTh) # for threshold expression in file name
# Target (Yellow) Mask Threshold
    V2 = str(abs(LTh / 255.0 * Lmax)) # actual Yellow mask Threshold
    IJ.run(imp1, "Subtract...","value=" + V2)
    if Mode == "32-bit": # clip value under 0.0 for 32bit image
        imp1 = clipUnder0(imp1, Lmax)
    if NegaInvert == True: # for making inverted Negative mask
        IJ.run(imp1, "Multiply...", "value=" + str(4 / TrFac))
            # making mask more transparent (x 4 for Nega)
        imp1.getProcessor().invert()
        V1 = V1 + "_Nega"
        V = "_Nega_x "+ str(TrFac)
    else: # for making positive mask
        IJ.run(imp1, "Multiply...", "value=" + str(4 * TrFac))
            # making mask more transparent (x 4)
        V1 = V1 + "_Posi"
        V = "_Posi_x" + str(TrFac)
    impLumMask = imp1
    if Show == "Yes":
        impLumMask.show()  # Show the subtracted image in window
    return impLumMask, V, V1
    """
    return values
        impGraymask: mask image
        V: description for mask parameters
        V1: description for mask parameters
    """
#-------------------------------------------
def make_binary_mask(imp, zoneLTh, zoneHTh, Mode, Show):
    """
    parameters
        imp: Original Ch.image
        impZone: Working  Ch.image
        zoneLTh: Lower threshold, zoneHTh: Upper threshold
        Mode: bit depth, Show: Show windows

    """
# make impZone for Grayscale 8 bit!!
    Lmax = set_Lmax(imp)
    impZone = Duplicator().run(imp)
    impZone.setDisplayRange(0.0, Lmax)
    # ImageConverter(impZone).setDoScaling(True)
#           it is needs for 16bit image to down convert to 8bit
    ip = impZone.getProcessor() # xxx
    if Mode == "8-bit":
        ip_mask = ByteProcessor(ip.width, ip.height)
    elif Mode == "16-bit":
        ip_mask = ShortProcessor(ip.width, ip.height)
    elif Mode == "32-bit":
        ip_mask = FloatProcessor(ip.width, ip.height)
    ip_mask.setValue(Lmax) # 0.53d
# make Zone ip_mask
    if zoneHTh >= 255:
        zoneHTh = 255
    impZone.getProcessor().setThreshold((zoneLTh / 255.0 * Lmax), \
        (zoneHTh / 255.0 * Lmax), ImageProcessor.NO_LUT_UPDATE)
    print ("Zone L & H", zoneLTh, zoneHTh)
    roi = ThresholdToSelection.run(impZone)
    ip_mask.setRoi(roi)
    ip_mask.fill(roi)
    try:
        impmask = ImagePlus("mask", ip_mask)
        impmask.setDisplayRange(0.0, Lmax)
#        Zone ip_mask made from ROI: Zone ip_mask: impip_mask
    except: # exception process for ROI missing
        IJ.error("Selected ROI (Region of Interest) is missing.")
        if Show == "Yes":
            impZone.hide()
        impZone.close()
        return
    if Show == "Yes":
        impmask.show() # show Zone ip_mask
    impZone.close()
    return impmask
    """
    return values
        impMask: mask image
    """


#----------------------------------------------
def adjMaskContrast(imp, Vcontrast, LocalMode):
    Lmax = set_Lmax(imp)
    # This routine is for adjustment mask contrast (Ver. 0.45)
    # imp: Mask image
    # Vcontrast: Contrast factor 0.0 to 1.0
    # LocalMode: image bit depth (in local)
    adjValue = Lmax - Lmax * Vcontrast
    imp.getProcessor().multiply(Vcontrast)
    imp.getProcessor().add(adjValue)
    return imp
#----------------------------------------------
def channelSelection(colorCh, impR, impG, impB, DoAdjLevel):
    # DoAdjLevel: Do level adjustment between channels?
    if colorCh == "Yellow/Blue" or colorCh == "Yellow":
        imp1 = Duplicator().run(impB)
        imp2 = Duplicator().run(impG)
        imp3 = Duplicator().run(impR)
    elif colorCh == "Cyan/Red" or colorCh == "Cyan":
        imp1 = Duplicator().run(impR)
        imp2 = Duplicator().run(impG)
        imp3 = Duplicator().run(impB)
    elif colorCh == "Magenta/Green" or colorCh == "Magenta":
        imp1 = Duplicator().run(impG)
        imp2 = Duplicator().run(impR)
        imp3 = Duplicator().run(impB)

    stat1 = imp1.getStatistics()
    stat2 = imp2.getStatistics()
    stat3 = imp3.getStatistics()
    if DoAdjLevel == True:
        IJ.run(imp2, "Add...", "value=" + str(stat1.mean - stat2.mean))
        IJ.run(imp3, "Add...", "value=" + str(stat1.mean - stat3.mean))
    return imp1, imp2, imp3
#-----------------------------------------------
def baseChannelSelection(baseCh, impLuminosity, impR, impG, impB):
    if baseCh == "Luminosity":
        impZone = Duplicator().run(impLuminosity)
        chName ="Gray"
        zoneLTh = 0 # for Luminosity mask, zone lower threshold is always 0
    elif baseCh == "Blue":
        impZone = Duplicator().run(impB)
        chName ="B"
    elif baseCh == "Red":
        impZone = Duplicator().run(impR)
        chName ="R"
    elif baseCh == "Green":
        impZone = Duplicator().run(impG)
        chName ="G"
    impZone.getProcessor().setLut(None)
    return impZone, chName

#----------------------------------------------
def printStat(Stat, string):
    print (string + " mean", Stat.mean)
    print (string + " min", Stat.min)
    print (string + " max", Stat.max)
    print (string + " mode", Stat.mode)

#----------------------------------------------
### RGB Manipulation
def RGBmanipulation(imp0, parm, stackSize):
# this is mask making & saving main routine
    """
    parameters
        imp0: original image (Hyperstack color image)
        folder: folder name of original image file
        filename: file name of original image file
        baseCh: channel name for deciding mask upper & lower thresholds
        parm: parameters array from main routine
        color Ch: channel name for making color mask image
        stackSize: Color file or MonoChrome / 1 is Mono 3 is Color
        sequence: Mask sequcence No.
    global varialbes
        Mode: bit depth (or process canselling)
        Lmax: maximum image lightness value
        imagePlus
    """
#  input Parameters
    Lmax = set_Lmax(imp0)
    if parm is not None: # if parm is not null, then get paraneters...
        Mode = parm[0] # Mode: Image bit depth
        # IJ.showMessage("NextMask" + str(NextMask))
        imp0.show()
    impLuminosity = convertToLuminosity(imp0)
                    # Luminosity Gray image
    impsCh = ChannelSplitter.split(imp0) #Channel Splitting of imp0

    for i in impsCh: # it needs to convert Grayscale image
        if Mode == "32-bit":
            ImageConverter(i).convertToGray32()
        elif Mode == "8-bit":
            ImageConverter(i).convertToGray8()
        elif Mode == "16-bit":
            ImageConverter(i).convertToGray16()
        i.setDisplayRange(0.0, Lmax)

    if stackSize!=1: #if imp0 is color then...
    ## R ch.
        impR =  impsCh[0]
    ## G ch.
        impG =  impsCh[1]
    ## B ch. & B x 0.3
        impB =  impsCh[2]
    imp0.hide()
    return impR, impG, impB, impLuminosity


def baseProsess(impR, impG, impB, impLuminosity, folder,filename, \
                            baseCh, parm, colorCh, stackSize,sequence):
    if parm is not None: # if parm is not null, then get paraneters...
        Mode, YTh, blInvert, zoneLTh, zoneHTh, TrFac, Show, Gaussian, \
            GaussianSigma, MaskInvert, NextMask, DoAdjLevel, imp2MixRatio,\
                                                    Vcontrast = parm
        """
        variables
            Mode: Image Format 16-bit or 8-bit
            YTh: yellow(or blue) value threshold of mask
            zoneLTh: lower threshold of mask
            zoneHTh: upper threshold of mask
            blinvert: make primary color (blue) mask
                                        (negative mask in Luminosity) or not
            Show: working wndows Showing Mode
            Gaussian: Apply or not Gaussian Blur
            GaussianSigma: Gaussian Blur Sigma
            MaskInvert: Mask Inversion for excluding selected area
                                                                (True / False)
            NextMask: Making Next Mask
            Vcontrast: Mask contrast adjustment value
        """
    # Show Windows
    if Show == "Yes":
        if stackSize ==3: #if imp0 is color then...
            impR.show()
            impG.show()
            impB.show()
        else:
            impLuminosity.show()
    # def chcannelSelection()
    if colorCh != "Luminosity": # 3 channels
        imp1, imp2, imp3 = \
            channelSelection(colorCh, impR, impG, impB, DoAdjLevel)
    elif colorCh == "Luminosity": # n of ch. is 1 or 3
        imp1 = Duplicator().run(impLuminosity)
        imp1.getProcessor().setLut(None)
        # No imp2 and imp3


# Grayscale Mask making process start
    if colorCh != "Luminosity": # not Luminosity (3 ch.)
#$$ Make relative RGB Color Mask
        impYmask, V = grayscale_mask_maker(imp1, imp2, imp3, YTh, Mode, \
                            blInvert, Show, colorCh, TrFac, imp2MixRatio)
        impYmask.setDisplayRange(0.0, Lmax)

#    save relative RGB color Mask
#     Mask inversion (add 0.15)
        if MaskInvert == False or sequence ==2: # Mask No Inversion
            # adjustment mask contrast  (Ver. 0.45)
            impYmask = adjMaskContrast(impYmask, Vcontrast, Mode)
            suffix = "_Mask_Thr_".decode('utf-8') + V.decode('utf-8') + \
                "_x4x" + str(TrFac)
                    # Save the Subtracted image x 4
        elif sequence ==1: # Mask Inversion for 1st mask
            IJ.run(impYmask, "Invert", "")
            # adjustment mask contrast  (Ver. 0.45)
            impYmask = adjMaskContrast(impYmask, Vcontrast, Mode)
            suffix =  "_Mask_Thr_" + V + "_x4x" + str(TrFac) + "_Inv"
        impYmask.setDisplayRange(0.0, Lmax)
        IJ.saveAsTiff(impYmask, folder + filename + suffix + ". tif")
                    # Save the Subtracted image x 4
    else:
#$$ Make Luminosity Mask (1 or 3 ch.)
        impYmask, V, V1 = luminosity_mask_maker(imp1, zoneLTh, Mode,\
                                                blInvert, Show, TrFac)
        impYmask.setDisplayRange(0.0, Lmax)

#    save Luminosity Mask
#     Mask inversion (add Ver. 0.15)
        if MaskInvert == False or sequence ==2: # Mask No Inversion
            # adjustment mask contrast (Ver. 0.45)
            impYmask = adjMaskContrast(impYmask, Vcontrast, Mode)
            suffix = "_Mask_Thr_" + "_x4x" + str(TrFac)
                    # Save the Subtracted image x 4 x transparency factor
        elif sequence == 1: # Mask Inversion for 1st mask
            IJ.run(impYmask, "Invert", "")
            # adjustment mask contrast (Ver. 0.45)
            impYmask = adjMaskContrast(impYmask, Vcontrast, Mode)
            suffix = "_Mask_Thr_" + V1 + "_x4x" + str(TrFac) + "_Inv"
        impYmask.setDisplayRange(0.0, Lmax)
        IJ.saveAsTiff(impYmask, folder + filename + suffix + ". tif")

                    # Save the Subtracted image x 4 x transparency factor
# Binary Mask Making process start
# select target channel for making lightness range binary mask
#     impZone: base channel image
# baseChannel Selection
    if stackSize == 3:
        impZone, chName = \
            baseChannelSelection(baseCh, impLuminosity, impR, impG, impB)
    else: # mono channel
        impZone = Duplicator().run(impLuminosity)
        chName = "Gray"
    impZone.setDisplayRange(0.0, Lmax)

    imp0.close()
#$$ Make binary mask for deciding mask valid lightness range
    if colorCh != "Luminosity":
        impMask = make_binary_mask(impZone, zoneLTh, zoneHTh, Mode, Show)
    else: # Luminosity mask
        impMask = make_binary_mask(impZone, 0.0, zoneHTh, Mode, Show)
        impMask = adjMaskContrast(impMask, Vcontrast, Mode)
        # in Luminosity mask lowert threshold = 0.0
    impMask.setDisplayRange(0.0, Lmax)

#      impMask may be 8bit or 16bit depending on original image bit
#   after deepen bit depth, Save Zone Mask
    suffix = "_" + chName + "_Zone" \
        + str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + "_Mask"
    IJ.saveAsTiff(impMask, folder + filename + suffix + ". tif")
#   --------------------------------------------------
#$$ bind valid lightness range binary mask and relative RGB color mask
    ic = plugin.ImageCalculator()
    maskBindimp = ic.run("Min created", impYmask, impMask) # bind masks
    maskBindimp.setDisplayRange(0.0, Lmax)

    if Show == "Yes":
        maskBindimp.show()
#   apply Gaussian blur
    if (Gaussian == True) and (colorCh !="Luminosity"):
        IJ.run(maskBindimp, "Gaussian Blur...", "sigma=" + str(GaussianSigma))
#   Mask Inversion for sequence 2
    if MaskInvert == True and sequence ==2:
        IJ.run(maskBindimp, "Invert", "")
##### save binded mask
    if sequence == 2 or NextMask == True:
        # saving file name for Binding 2 Masks
        suffix = "_tmpMask_" + str(sequence)
                # Save the Subtracted image x 4
    else: # saving file name for 1 Mask only
        if MaskInvert == False: # Mask No Inversion
            suffix = "_" + chName + \
                "_Zone" + str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + \
                "_" + V + "_Mask_Binded"

        else: # Mask Inversion
            suffix = "_" + chName + \
                "_Zone" + str(int(zoneLTh)) + "-" + str(int(zoneHTh)) + \
                "_" + V + "_Mask_Binded_and_Inverted"
    if Mode == "32-bit":
        maskBindimp.setDisplayRange(0.0, Lmax)

    IJ.saveAsTiff(maskBindimp, folder + filename + suffix + ". tif")

#     close working windows
    if Show == "Yes":
        impYmask.hide()
        impZone.hide()
        impMask.hide()
        maskBindimp.hide()
    impYmask.close()
    impMask.close()
    maskBindimp.close()
    impZone.close()
#   End of def RGBmanipulation
#----------------------------------------------
class ThresholdPreviewer(AdjustmentListener, ItemListener):
    # previewer for binary mask for RGB color mask 
    #               and observe changes of dialog items
    def __init__(self, impMainWin, imp, nfield_slider_upper, \
                 nfield_slider_lower, checkbox_prv, checkbox_finalprv):
        """
            imp*: an ImagePlus
            slider_*: a java.awt.Scrollbar UI element
            checkbox_*: a java.awt.Checkbox controlling whether to
                                    dynamically update the ImagePlus as the
                                    scrollbar is updated, or not.
            imp: preview image
            slider_upper: slider for upper lightness threshold
            slider_lower: slider for lower lightness threshold
            checkbox_prv: checkbox for preview image on / off
            checkbox_finalprv: checkbox for showing final mask image
        """
        self.imp = imp
        self.impOrg = Duplicator().run(self.imp)
        self.imp.setTitle("valid lightness range preview")
        self.impOrg.setTitle("Original preview imp")
        self.original_ip = imp.getProcessor().duplicate() # store a copy
        self.impMainWin = impMainWin
        self.impMainWin.setTitle("main preview windows")
        self.nfield_slider_upper = nfield_slider_upper
        self.nfield_slider_lower = nfield_slider_lower
        self.checkbox_prv = checkbox_prv
        self.checkbox_finalprv = checkbox_finalprv

    def adjustmentValueChanged(self, event):
        """ event: an AdjustmentEvent with data on 
                            the state of the scroll bar. """
        preview = self.checkbox_prv.getState()
        if preview:

            if (IJ.isMacOSX() == False) and event.getValueIsAdjusting():
                return
                    # there are more scrollbar adjustment events queued already
            self.checkbox_finalprv.setState(False)
            self.threshold()

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on 
                            what happened to the checkbox. """
        if event.getStateChange() == event.SELECTED:
            self.checkbox_finalprv.setState(False)
            self.threshold()

    def reset(self):
        """ Restore the original ImageProcessor """
        self.imp.setProcessor(self.original_ip)

    def threshold(self):
        """ Execute the in-place making binary image of the ImagePlus. """
        threshold_upper = float(self.nfield_slider_upper.getText())
        threshold_lower = float(self.nfield_slider_lower.getText())
        if threshold_upper < threshold_lower:
            threshold_upper = threshold_lower
        # IJ.showMessage("Threshold L & H: " + str(threshold_lower) + " " + str(threshold_upper))
        impBinalyMask = make_binary_mask(self.impOrg, threshold_lower, \
                                         threshold_upper, "8-bit", "No")
        ip_mask = impBinalyMask.getProcessor()
        # copy the image to showing windows
        self.impMainWin.setProcessor(ip_mask)
        self.impMainWin.getProcessor().setLut (None)
        # keep the image for final preview image
        self.imp.setProcessor(ip_mask)
        self.imp.getProcessor().setLut (None)
        self.imp.hide()
        self.impMainWin.show()
        IJ.selectWindow(self.impMainWin.getTitle())
        WM.setWindow(self.impMainWin.getWindow())
        impBinalyMask.close()

#----------------------------------------------
class ColorMaskPreviwer(ItemListener, TextListener, AdjustmentListener):
    # Previwer for RGB Color grayscale mask and observe changes of dialog items
    def __init__(self, impMainWin, impPrv, impPrvR, impPrvG, impPrvB, \
        C0, C1, nfield_maskThr, nfield_TrFac, checkbox_prv, \
        chekbox_PrNgMask, checkbox_inverse, nfield_slider_lower, \
        checkbox_adjlevel, checkbox_finalprv, sequence, \
                                    slider_imp2MixRatio , nfield_contrast):
        """
            imp*: an ImagePlus
            slider_*: a java.awt.Scrollbar UI element
            preview_*: a java.awt.Checkbox controlling whether to
                                    dynamically update the ImagePlus as the
                                                scrollbar is updated, or not.
            nfield_*: a java.awt.TextField
            impPrv: preview image handler,
            impPrvR: R ch. / impPrvG: G ch. / impPrvB: B ch.
            C0: complimentary color name of target ch.
            C1: primary color name of target ch.
            nfield_maskThr: numeric fileld of color threshold
            nfield_TrFac: numeric fileld of transparency factor
            checkbox_prv: checkbox for preview image on/off (True/False)
            chekbox_PrNgMask: checkbox for making primary color mask
            checkbox_inverse:  checkbox for making excluding mask
            nfield_slider_lower: deleted
            checkbox_adjlevel: checkbox for executing level
                                    adjustment between channels
            checkbox_finalprv: checkbox for showing final mask image On/Off
            slider_contrast: slider for mask contrast adjunstment
            nfield_contrast: nfield for mask contrast adjunstment
            checkbox_Rel2B: checkbox for refering only B ch.
                                                (only in making G/Mg mask)
        """
        self.impPrv = impPrv
        self.impPrv.setTitle("relative color mask preview")
        self.impMainWin = impMainWin
        self.impMainWin.setTitle("main preview windows")
        self.impPrvR = impPrvR
        self.impPrvG = impPrvG
        self.impPrvB = impPrvB

        statR = impPrvR.getStatistics()
        statG = impPrvG.getStatistics()
        statB = impPrvB.getStatistics()

        self.C0 = C0
        self.C1 = C1
        self.C = C0
        self.checkbox_prv = checkbox_prv
        self.chekbox_PrNgMask = chekbox_PrNgMask
        self.checkbox_inverse = checkbox_inverse
        self.checkbox_adjlevel = checkbox_adjlevel
        self.checkbox_finalprv = checkbox_finalprv
        self.nfield_maskThr = nfield_maskThr
        self.nfield_TrFac = nfield_TrFac
        self.nfield_slider_lower = nfield_slider_lower
        self.sequence = sequence
        self.nfield_contrast = nfield_contrast
        self.original_ip = self.impPrv.getProcessor().duplicate()
#       ----for research version---
        self.slider_imp2MixRatio = slider_imp2MixRatio
#       ---------------------------
        self.statR = statR
        self.statG = statG
        self.statB = statB

    def textValueChanged(self, event):
        """ event: a textValueChanged with data on the state of 
                                                the numeric field. """
        try:
            self.textStrings = self.nfield_maskThr.getText() 
                                        # threshold for mask
            self.textStrings2 = self.nfield_TrFac.getText() 
                                        # Transparency factor for mask

        except:
            return
        self.checkbox_finalprv.setState(False)
        preview = self.checkbox_prv.getState()
        if preview:
            if self.C0 != "Luminosity Positive":
                self.threshold()
        return

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on 
                                        what happened to the checkbox. """
        self.checkbox_finalprv.setState(False)
        preview = self.checkbox_prv.getState()
        if preview:
            if self.C0 != "Luminosity Positive":
                self.threshold()

    def adjustmentValueChanged(self, event):
        """ event: an AdjustmentEvent with data on 
                                            the state of the scroll bar. """
        preview = self.checkbox_prv.getState()
        if preview:

            if (IJ.isMacOSX() == False) and event.getValueIsAdjusting():
                return
                    #there are more scrollbar adjustment events queued already
            self.checkbox_finalprv.setState(False)
            self.threshold()


    def reset(self):
        """ Restore the original ImageProcessor """
        self.impPrv.setProcessor(self.original_ip)

    def threshold(self):
        """ Execute the in-place making binary image of the ImagePlus. """
#       ----for research version---
        Bmask = self.chekbox_PrNgMask.getState()
                # Bmask value: making primary color mask? True/False
        AdjLevel = self.checkbox_adjlevel.getState()
                # Do channel level adjustment?

        impPrv0, impPrv1, impPrv2 = channelSelection\
            (self.C0, self.impPrvR, self.impPrvG, self.impPrvB, AdjLevel)

        MixRatio = float(self.slider_imp2MixRatio.getValue()) / 100
        if Bmask == False: # for making complementary color mask
            MixRatio = 1.0 - MixRatio
        if MixRatio > 1.0:
            MixRatio = 1.0
        elif MixRatio < 0.0:
            MixRatio = 0.0

        # difinision of including / excluding mask
        self.Inverse = self.checkbox_inverse.getState()
        # getting numeric parameters
        self.textStrings = self.nfield_maskThr.getText() 
                            #  field for color threshold
        self.textStrings2 = self.nfield_TrFac.getText()
                            #  field for transparency factor for mask
        self.maskThr = int(self.textStrings) # threshold for mask
        self.maskTrFac = float(self.textStrings2)
                                        # Transparency factor for mask
        self.maskContrast = float(self.nfield_contrast.getText()) / 100
                                        # Contrast factor for mask
        # difinition of complementary or primary color mask
        if Bmask == True: # primaly color mask
            self.C = self.C1
        else: # complementary color mask
            self.C = self.C0

        impPrvMask, V = grayscale_mask_maker(impPrv0, impPrv1, \
            impPrv2, self.maskThr, "8-bit", Bmask, "No", \
            self.C, self.maskTrFac, MixRatio)
            # Mode: 8-bit, Show: "No"

        if self.Inverse == True: # make excluding mask preview
            IJ.run(impPrvMask, "Invert","")
        impPrvMask = \
            adjMaskContrast(impPrvMask, self.maskContrast, "8-bit")
                        # mask contrast adjustment
        ip_PrvMask = impPrvMask.getProcessor()
        ip_PrvMask.BLACK_AND_WHITE_LUT
        # copy the image to showing windows
        self.impMainWin.setProcessor(ip_PrvMask)
        self.impMainWin.getProcessor().setLut (None)
        self.impMainWin.show()
        # keep the image for final preview image
        self.impPrv.setProcessor(ip_PrvMask)
        self.impPrv.getProcessor().setLut (None)
        self.impPrv.hide()
        # close temporary working imagePlus
        # impPrvTmp.close()
        impPrvMask.close()
        impPrv0.close()
        impPrv1.close()
        impPrv2.close()
        IJ.selectWindow(self.impMainWin.getTitle())
        WM.setWindow(self.impMainWin.getWindow())

#-------------------------------------------
class LuminosityMaskPreviewer(AdjustmentListener, ItemListener, TextListener):
        # previewer for binary & grayscale mask of luminosity
        #                      and observe changes of dialog items
    def __init__(self, impPrv, nfield_slider_upper, nfield_slider_lower, \
        C0, C1, nfield_TrFac, checkbox_prv, chekbox_PrNgMask, \
                nfield_contrast):
        """
            imp*: an ImagePlus
            slider_*: a java.awt.Scrollbar UI element
            checkbox_*: a java.awt.Checkbox controlling whether to
                                    dynamically update the ImagePlus as the
                                    scrollbar is updated, or not.
            impCanvas: handler for preview image
            slider_upper: slider for upper lightness threshold
            slider_lower: slider for lower lightness threshold
            C1: mask name for Negative Mask / C0: mask name for Positive Mask
            nfield_TrFac: Numeric field for transparency factor
            checkbox_prv: checkbox for preview image on / off
            chekbox_PrNgMask: making Negative(True) / Positive(False) mask
            checkbox_L_range: changing preview image
                True: Binary Mask  False: Grayscale mask
            checkbox_finalprv: checkbox for showing final mask image
            slider_contrast:
            nfield_contrast: nfield for value of contrast adjustment
        """
        self.impPrv0 = Duplicator().run(impPrv)
        self.impCanvas= impPrv
        self.impCanvas.setTitle("luminosity mask preview")
        impCanvasStat = self.impCanvas.getStatistics()
        printStat(impCanvasStat, "impCanvas")
        self.impPrv0.setTitle("impPrv0")
        self.impPrv0.hide()
        self.original_ip = impPrv.getProcessor().duplicate() # store a copy
        self.nfield_slider_upper = nfield_slider_upper
        self.nfield_slider_lower = nfield_slider_lower
        self.C0 = C0
        self.C1 = C1
        self.C = C0
        self.checkbox_prv = checkbox_prv # Preview checkbox
        self.chekbox_PrNgMask = chekbox_PrNgMask 
                        # Negative / Positive checkbox
            #Attention! use it for exchanging preview image in Luminosity mask
        self.nfield_TrFac = nfield_TrFac # transparency factor
        textStrings2 = self.nfield_TrFac.getText()
                            #  field for transparency factor for mask
        self.maskTrFac =  float(textStrings2) # transparency factor for mask

        self.nfield_contrast = nfield_contrast
        ip = impPrv.getProcessor()
        self.impMask = ImagePlus("Mask Dummy", ip)
        self.impLumPrv = ImagePlus("Luminosity Dummy",ip)
        self.NegaInvert = self.chekbox_PrNgMask.getState()
                        # check box for selecting negative / positive image

    def adjustmentValueChanged(self, event):
        """ event: an AdjustmentEvent with data on 
                                    the state of the scroll bar. """
        preview = self.checkbox_prv.getState()
        if preview:

            if (IJ.isMacOSX() == False) and event.getValueIsAdjusting():
                return
                # there are more scrollbar adjustment events queued already
            #self.threshold()
            self.luminosity_prv() # make grayscale preview

    def textValueChanged(self, event):
        """ event: a textValueChanged with data on 
                                the state of the numeric field. """
        try:
            textStrings2 = self.nfield_TrFac.getText()
            self.maskTrFac = float(textStrings2) 
                            # Transparency factor for mask
        except:
            return
        preview = self.checkbox_prv.getState()
        if preview:
            self.luminosity_prv()
        return

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on 
                                what happened to the checkbox. """
        self.NegaInvert = self.chekbox_PrNgMask.getState()
                    # check box for selecting negative / positive image
        self.luminosity_prv()

    def reset(self):
        """ Restore the original ImageProcessor """
        self.impCanvas.setProcessor(self.original_ip)

    def luminosity_prv(self):
        """ Execute the in-place making luminosity image of the ImagePlus. """
        threshold_upper = int(self.nfield_slider_upper.getText())
        threshold_lower = float(self.nfield_slider_lower.getText())

        self.impLumPrv.close()
        impLumPrv, V, V1 = luminosity_mask_maker(self.impPrv0, \
            threshold_lower, "8-bit", self.NegaInvert, "No", self.maskTrFac)
        self.impMask.close()
        impMask = make_binary_mask(self.impPrv0, 0.0, \
            threshold_upper, "8-bit", "No")  # threshold_lower to 0
        ic = plugin.ImageCalculator()
        impFinalPrv = ic.run("Min created", impLumPrv, impMask)
        ip_PrvMask = impFinalPrv.getProcessor()
        self.impCanvas.setProcessor(ip_PrvMask)
        self.impCanvas.getProcessor().setLut (None)
        self.maskContrast = float(self.nfield_contrast.getText()) / 100
        self.impCanvas = \
            adjMaskContrast(self.impCanvas, self.maskContrast, "8-bit")
                            # mask contrast adjustment
        self.impCanvas.updateAndDraw()
        IJ.selectWindow(self.impCanvas.getTitle())
        WM.setWindow(self.impCanvas.getWindow())
        self.impLumPrv = impLumPrv
        self.impMask = impMask
        impFinalPrv.close()

#---------------------- 0.27
class FinalImagePreviewer(ItemListener):
    """ previewer for binary mask for RGB color mask 
                                and observe changes of dialog items """
    def __init__(self, impMainWin, impBin, impGray, checkbox_finalprv, \
        checkbox_inverse, sequence):
        """
            impPrv, impBin, impGray: an ImagePlus
            impBin: Binary mask image / impGray: Grayscale mask image
            impPrv: Preview window handler
            slider: a java.awt.Scrollbar UI element
            checkbox_finalprv: a java.awt.Checkbox controlling whether to
                                    dynamically update the ImagePlus as the
                                    scrollbar is updated, or not.
            Inverse2nd: in case of 2nd mask and image inverted for excluding
        """
        self.impMainWin  = impMainWin
        impMainWin.setTitle("* final image preview")
        self.impBin  = impBin
        self.impGray = impGray
        self.checkbox_finalprv = checkbox_finalprv
        self.checkbox_inverse = checkbox_inverse
        self.sequence = sequence

    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on 
                                        what happened to the checkbox. """
        if event.getStateChange() == event.SELECTED:
            self.combineImage()

    def combineImage(self):
        """ combine two preview image to final image preview. """
        impGrayTmp = Duplicator().run(self.impGray)
        self.Inverse2nd = self.checkbox_inverse.getState()
        # whether this turn is 2nd sequence and making excluding mask...
        if self.sequence == 1:
            self.Inverse2nd = False
        if self.Inverse2nd == True:
            IJ.run(impGrayTmp, "Invert","")
        ic = plugin.ImageCalculator()
        impFinalPrv = ic.run("Min created", self.impBin , impGrayTmp)
        self.impMainWin.setProcessor(impFinalPrv.getProcessor())
        impGrayTmp.close()
        # set Final mask image to impMainWin
        if self.Inverse2nd == True:
            IJ.run(self.impMainWin, "Invert","")
        self.impMainWin.show()
        IJ.selectWindow(self.impMainWin.getTitle())
        WM.setWindow(self.impMainWin.getWindow())
        self.checkbox_finalprv.setState(False)
        # Show Final mask image window in front

#----- dialog for RGB Color Mask-------------------------------------------
def thresholdImageUIa(imp, impPrvR, impPrvG, impPrvB, Mode, C0, C1, sequence):
    """ imp (preview image) is already 8-bit monochannel """
    impStat = imp.getStatistics()
    printStat(impStat, "impStat")
    imp.show()
    WM.getActiveWindow().toFront()
    gd = GenericDialog(C0 +" (" + C1 + \
                                ") Mask Maker / input threshold parameters")
    #----- Dialog parts arrangement -------

    gd.setInsets(0, 0, 0)
    gd.addMessage("[ Relative " + C0 +" (" + C1 + ") mask parameters ]")
    gd.setInsets(0, 20, 0)
    gd.addMessage("------ Color parameters----")
    gd.setInsets(0, 20, 0)
    if C0 == "Cyan" and sequence == 2:  # make second red mask
        gd.addCheckbox("Make " + C1 + " (primary color) Mask", True) 
                                                                #ChkBox 0
    else: # make normal RGB mask
        gd.addCheckbox("Make " + C1 + " (primary color) Mask", False) 
                                                                #ChkBox 0
    if C0 == "Magenta":
        defaultValue = 25
    else:
        defaultValue = 10
    gd.setInsets(0, 0, 0)
    gd.addNumericField("Color Threshold", defaultValue, 0) #NField 0
    gd.addToSameRow()
    gd.addMessage(" (Signed relative value)")

    gd.addMessage("Skew of color sensitivity")

    #NField 6 comments
    if C0 == "Yellow":
        imp2name = "R"
        imp3name = "G"
    elif C0 == "Cyan":
        imp2name = "G"
        imp3name = "B"
    elif C0 == "Magenta":
        imp2name = "R"
        imp3name = "B"
    gd.setInsets(0, 0, 0)
    gd.addSlider("", 0.0, 1.0, 0.5, 0.01)  #slider 0 / NField 1
    gd.addToSameRow()
    gd.addMessage(imp2name + "(0.0) ---- Neutral(0.5) ---- (1.0)" + \
                                imp3name)

    gd.addCheckbox("Adjust level between channels", False) #ChkBox 1

    gd.setInsets(0, 20, 0)
    gd.addMessage("------ transparency / contrast parameters----")
    gd.addNumericField("Mask Transparency", 1.0, 1)  #NField 2
    gd.addToSameRow()
    gd.addMessage("to make binary mask instead of grayscale mask, \n"\
        + "  set Mask Transparency Factor = 100")
    gd.setInsets(0, 0, 0)
    gd.addSlider("Mask Contrast", 0, 100, 100)  #slider 1 / NField 3

    if C0 == "Cyan" and sequence == 2: # make second red mask
        gd.addCheckbox\
            ("Mask Inversion for Excluding selected area", True)
                                                        #ChkBox 2 (Mg)
    else: # make normal RGB mask
        gd.addCheckbox\
            ("Mask Inversion for Excluding selected area", False)
                                                        #ChkBox 2 (RGB)
    gd.setInsets(0, 20, 0)
    gd.addCheckbox("Preview", True) #ChkBox 3

    gd.addCheckbox("Apply Gaussian Blur", False) #ChkBox 4 (RGB)
    gd.addToSameRow()
    gd.addNumericField("Gaussian Blur Sigma", 2.0, 1)  #NField 4 (RGB)
    gd.setInsets(5, 0, 0)
    gd.addMessage("[ Effective tone area selection ]")
    gd.setInsets(2, 0, 0)
    gd.addMessage("Mask effective L range (0-255)\n" + \
                "        White parts in preview are mask effective areas")
    gd.setInsets(0, 0, 0)
    gd.addSlider("Upper Threshold", 0, 255, 128)  #slider 2 / NField 5
    gd.addSlider("Lower Threshold", 0, 255, 128)  #slider 3 / NField 6

    gd.addCheckbox("Show final image preview", False) # 0.27 #ChkBox 5 (RGB)

    gd.addMessage("--------------------------------------")
    if sequence ==1:
        gd.addCheckbox("Make combined Mask", False) #ChkBox 6
    showWin = ["Yes","No"]
    gd.setInsets(0, 30, 0)
    gd.addChoice("Show working windows?", showWin, showWin[0]) # choice 0

    #------ The UI elements for the above inputs ------
    chekbox_PrNgMask = gd.getCheckboxes().get(0) # CBox #0
        #  chkbox for primary color mask / or negative mask (Luminosity)
    nfield_maskThr = gd.getNumericFields().get(0) # NField #0: Color Thr
    # skew of color sensitivity
    slider_imp2MixRatio  = gd.getSliders().get(0) # slider #0 (RGB)
    nfield_imp2MixRatio  = gd.getNumericFields().get(1) # NField #1 (RGB)
    checkbox_adjlevel = gd.getCheckboxes().get(1)  
                                    # CBox #1 channel level adjustment
    checkbox_inverse = gd.getCheckboxes().get(2) 
                                    # CBox #2 Inversion for excluding mask
    nfield_TrFac = gd.getNumericFields().get(2) 
                                    # NField #2 Transparency Factor
    slider_contrast = gd.getSliders().get(1) # slider #1 mask contrast slider
    nfield_contrast = gd.getNumericFields().get(3) 
                                            # NField #3 mask contrast value

    checkbox_prv = gd.getCheckboxes().get(3) # CBox #3 chkbox for preview
    # Gaussian blur # Cbox #4 and Nfield #4 (skip)
    #-----------------------------
    slider_upper = gd.getSliders().get(2) # slider #2 higher threshold
    slider_lower = gd.getSliders().get(3) #  slider #3 lower threshold
    nfield_slider_upper = gd.getNumericFields().get(5) 
                                                # NField #5 higher threshold
    nfield_slider_lower = gd.getNumericFields().get(6) 
                                                # NField #6 lower threshold
        # Attention! sliders also have NumericFields
    checkbox_finalprv =  gd.getCheckboxes().get(5) # CBox #5 (RGB)
                                # ChkBox final image preview
    #-----------------------------
    # Make Combined Mask ? # CBox #5 (Skip)
    choice_showWin = gd.getChoices().get(0) # show the process?

    """
            in Luminosity mask,
            checkbox_L_range & chekbox_PrNgMask have same function
                                                    (make negative mask)
    """

    if not imp:
        print ("Open an image first!")
        return
    # add event listeners
    #--add previewer1--- for binary mask
    # imp: handler for preview iamge of binary mask image
    imp.hide()
    impcp1 = Duplicator().run(imp)
    impcp1.hide()
    # impcp1: handler for preview iamge of grayscale mask image
    impcp2 = Duplicator().run(imp)
    impcp1.hide()
    # impcp2: handler for preview iamge of final mask image
    #    impcp2 will be used as only windows handler. 
    #       it's image will not be used.
    #----- set Previewers -----
    #-- Add RGB color binary previewer
    impcp1.show()
    previewer1 = ThresholdPreviewer(imp, impcp1, nfield_slider_upper, \
                nfield_slider_lower, checkbox_prv, checkbox_finalprv)
    slider_upper.addAdjustmentListener(previewer1) 
                                    # set listener for upper slider
    slider_lower.addAdjustmentListener(previewer1) 
                                    # set listener for lower slider
    checkbox_prv.addItemListener(previewer1)  
                                    # set listener for preview checkbox
    #--add previewer2--- for grayscale mask
    imp.hide()

    impcp2.show()
    # RGB color grayscale preview
    previewer2 = ColorMaskPreviwer(imp, impcp2, impPrvR, impPrvG, \
        impPrvB, C0, C1, nfield_maskThr, nfield_TrFac, checkbox_prv, \
        chekbox_PrNgMask, checkbox_inverse, nfield_slider_lower, \
        checkbox_adjlevel, checkbox_finalprv, sequence, \
        slider_imp2MixRatio, nfield_contrast)
    chekbox_PrNgMask.addItemListener(previewer2)
    checkbox_prv.addItemListener(previewer2)
    nfield_maskThr.addTextListener(previewer2)
    nfield_TrFac.addTextListener(previewer2)
    slider_contrast.addAdjustmentListener(previewer2)
    checkbox_inverse.addItemListener(previewer2)
    checkbox_adjlevel.addItemListener(previewer2)
    slider_imp2MixRatio.addAdjustmentListener(previewer2)

    previewerFinal = FinalImagePreviewer(imp, impcp1, impcp2, \
        checkbox_finalprv, checkbox_inverse, sequence)
        # imp: preview window handler / impcp1: binary mask image
        # impcp2: grayscale mask image
    checkbox_finalprv.addItemListener(previewerFinal)
#------  show daialog  ---------
    gd.showDialog()

    if gd.wasCanceled():
        previewer1.reset()
        previewer2.reset()
        Mode = "Stop"
        print ("User canceled dialog!")
#--------
#   input parameters
    blInvert = gd.getNextBoolean() 
            # ChkBox 0: Primary / Complementary Color Mask
    YTh =  gd.getNextNumber() # NField 0: Color Threshold

    imp2MixRatio =  float(gd.getNextNumber()) 
                    # NField 1: imp2 & 3 mix ratio (skew of c)
    if imp2MixRatio > 1.0:
        imp2MixRatio =  1.0
    elif imp2MixRatio < 0.0:
        imp2MixRatio =  0.0
    if blInvert == False: # for making complementary color mask
        imp2MixRatio = 1.0 - imp2MixRatio

    DoAdjLevel =  gd.getNextBoolean() 
            # ChkBox 1: Adjust level between channels
    MaskInvert =  gd.getNextBoolean()  
            # ChkBox 2: Invert for making excluding mask
#----------
    TrFac =  gd.getNextNumber() # NField 2: Transparency Factor

    Vcontrast =  gd.getNextNumber() # NField 3: Mask contrast
    Vcontrast = Vcontrast / 100

    preview_on = gd.getNextBoolean() # ChkBox 3: preview on / off (dummy)

    Gaussian = gd.getNextBoolean() # ChkBox 4: DO Gaussian blur?
    GaussianSigma =  gd.getNextNumber() # NField 4: Gaussian blur sigma
#----------
    zoneHTh =  gd.getNextNumber() # NField 5: effective range higher threshold
    zoneLTh =  gd.getNextNumber() # NField 6: effective range lower threshold

    tmp =  gd.getNextBoolean()  # ChkBox 5: final preview > dummy var

    if sequence ==1: # 1st dialog
        NextMask =  gd.getNextBoolean() # ChkBox 6: make combined mask?
    else: # 2nd dialog
        NextMask =  False
    Show =  gd.getNextChoice() # Choice 0: show converting process?
#       ------------------------------
    # all preview images will be closed
    #--binaly preview image
    imp.hide()
    imp.close()
    #--grayscale preview image (rgb color & luminosity)
    impcp1.hide()
    impcp1.close()
    impcp2.hide()
    impcp2.close()

    # remove event listeners
    #--remove previewer1--- for binary mask
    slider_upper.removeAdjustmentListener(previewer1) 
                                    # set listener for upper slider
    slider_lower.removeAdjustmentListener(previewer1) 
                                    # set listener for lower slider
    checkbox_prv.removeItemListener(previewer1)  
                                    # set listener for preview checkbox
#--remove previewer2--- for grayscale mask
    chekbox_PrNgMask.removeItemListener(previewer2)
    checkbox_prv.removeItemListener(previewer2)
    nfield_maskThr.removeTextListener(previewer2)
    nfield_TrFac.removeTextListener(previewer2)
    checkbox_inverse.removeItemListener(previewer2)
    slider_contrast.removeAdjustmentListener(previewer2)
    checkbox_inverse.removeItemListener(previewer2)
    checkbox_adjlevel.removeItemListener(previewer2)
    slider_imp2MixRatio.removeAdjustmentListener(previewer2)

    #--remove previewerL--- for Luminosity mask
    return Mode, YTh, blInvert, zoneLTh, zoneHTh, TrFac, str(Show), Gaussian,\
        GaussianSigma, MaskInvert, NextMask, DoAdjLevel, imp2MixRatio, \
        Vcontrast
    """
    return values
        YTh: Target Color Ch Threshold (originally Yellow Threshold)
        zoneLTh: lower threshold of mask
        zoneHTh: upper threshold of mask
        blInvert: Color invert Yes or No (originally invert for blue mask)
        Show: working wndows Showing Mode
        TrFac: Transparency Factor
        Gaussian: Apply or not Gaussian Blur
        GaussianSigma: Gaussian Blur Sigma
        MaskInvert: Mask Inversion for excluding selected area (True / False)
        DoAdjLevel: adjust level between channels
        Rel2B: compare to B only mode option (True/False)
    """
#----- dialog for Luminosity Mask --------------------------------------------
def thresholdImageUIb(imp, Mode, C0, C1, sequence, stackSize):
    """ imp (preview image) is already 8-bit monochannel """
    impStat = imp.getStatistics()
    printStat(impStat, "impStat")
    imp.show()
    WM.getActiveWindow().toFront()
    gd = GenericDialog(C0 +" (" + C1 + \
                                ") Mask Maker / input threshold parameters")
    #----- Dialog parts arrangement -------
    gd.addCheckbox("Make " + C1 + " Mask", False)  #ChkBox 0
    gd.addMessage("Select mask effective luminosity area (0-255)\n"+ \
                        "         White parts in preview are mask preview")
    gd.addSlider("Upper Threshold", 0, 255, 128)  #NField 0
    gd.addSlider("Lower Threshold (mask ref. point)", 0, 255, 128)  #NField 1
    gd.addCheckbox("Preview", True) #ChkBox 1
    gd.addNumericField("Mask Transparency Factor", 1.0, 1)  #NField 2
    gd.addSlider("Mask Contrast Adjustment", 0, 100, 100)  #NField 3
    showWin = ["Yes","No"]
    gd.addChoice("Show working windows?", showWin, showWin[0])

    if sequence == 1 and stackSize == 3:
        gd.addCheckbox("Make combined Mask", False) #ChkBox 2


    #------ The UI elements for the above inputs ------
    slider_upper = gd.getSliders().get(0) # slider #0 higher threshold
    slider_lower = gd.getSliders().get(1) #  slider #1 lower threshold
    chekbox_PrNgMask = gd.getCheckboxes().get(0) # CBox #0
        #  chkbox for primary color mask / or negative mask (Luminosity)
    checkbox_prv = gd.getCheckboxes().get(1) # CBox #1 chkbox for preview
    nfield_slider_upper = gd.getNumericFields().get(0) 
                                            # NField #0 higher threshold
    nfield_slider_lower = gd.getNumericFields().get(1) 
                                            # NField #1 lower threshold
        # Attention! sliders also have NumericFields
    nfield_TrFac = gd.getNumericFields().get(2) 
                                        # NField #2 Transparency Factor
    slider_contrast = gd.getSliders().get(2) 
                                        # slider #2 mask contrast slider
    nfield_contrast = gd.getNumericFields().get(3) 
                                        # NField #3 mask contrast value
    choice_showWin = gd.getChoices().get(0) # show the process?
    """
            in Luminosity mask,
            checkbox_L_range & chekbox_PrNgMask have same function
                                                    (make negative mask)
    """


    if not imp:
        print ("Open an image first!")
        return
    # add event listeners
    #--prepare preview images---
    # imp: handler for preview iamge of binary mask image
    imp.hide()
    impcp1 = Duplicator().run(imp)
    impcp1.hide()
    # impcp1: handler for preview iamge of grayscale mask image
    impcp2 = Duplicator().run(imp)
    impcp1.hide()
    # impcp2: handler for preview iamge of final mask image
    #    impcp2 will be used as only windows handler. 
    #       it's image will not be used.
    imp.hide()
    impcp2.show()
    #--set previewerL----------------
    # Luminosity mask preview (binary & grayscale)
    previewerL = LuminosityMaskPreviewer(impcp2, nfield_slider_upper, \
        nfield_slider_lower, C0, C1, nfield_TrFac, checkbox_prv, \
        chekbox_PrNgMask, nfield_contrast)
    slider_upper.addAdjustmentListener(previewerL) 
                                            # set listener for upper slider
    slider_lower.addAdjustmentListener(previewerL) 
                                            # set listener for lower slider
    chekbox_PrNgMask.addItemListener(previewerL)
    checkbox_prv.addItemListener(previewerL)  
                                        # set listener for preview checkbox
    nfield_TrFac.addTextListener(previewerL)
        # set listener for transparency fac. checkbox b
    slider_contrast.addAdjustmentListener(previewerL)
                                        # set listener for lower slider
        # in here use it for exchanging preview image
#------  show daialog  ---------
    gd.showDialog()

    if gd.wasCanceled():
        previewerL.reset()
        Mode = "Stop"
        print ("User canceled dialog!")
#   input parameters
    YTh =  0 # NField 0
    blInvert = gd.getNextBoolean() # ChkBox 0
    zoneHTh =  gd.getNextNumber() # NField 0
    zoneLTh =  gd.getNextNumber() # NField 1
    preview_on = gd.getNextBoolean() # ChkBox 1
    TrFac =  gd.getNextNumber() # NField 2
    Vcontrast =  gd.getNextNumber() # NField 3
    Vcontrast = Vcontrast / 100

    Gaussian = False
    GaussianSigma =  0.0
    Show =  gd.getNextChoice()
    MaskInvert =  False
    DoAdjLevel = False
    imp2MixRatio =  0.5
    if sequence == 1 and stackSize == 3: # 1st dialog
        NextMask =  gd.getNextBoolean() # ChkBox 2
    else: # 2nd dialog
        NextMask =  False
    # all preview images will be closed
    #--binaly preview image
    imp.hide()
    imp.close()
    #--grayscale preview image (rgb color & luminosity)
    impcp1.hide()
    impcp1.close()
    impcp2.hide()
    impcp2.close()

    # remove event listeners
    #--remove previewerL--- for Luminosity mask
    slider_upper.removeAdjustmentListener(previewerL)
        # remove listener for upper slider
    slider_lower.removeAdjustmentListener(previewerL)
        # remove listener for lower slider
    slider_contrast.removeAdjustmentListener(previewerL)
        # remove listener for contrast slider
    checkbox_prv.removeItemListener(previewerL)
        # remove listener for preview checkbox
    nfield_TrFac.removeTextListener(previewerL)
        # remove listener for transparency fac. checkbox
    return Mode, YTh, blInvert, zoneLTh, zoneHTh, TrFac, \
        str(Show), Gaussian, GaussianSigma, MaskInvert, \
        NextMask, DoAdjLevel, imp2MixRatio, Vcontrast
    """
    return values
        YTh: Target Color Ch Threshold (originally Yellow Threshold)
        zoneLTh: lower threshold of mask
        zoneHTh: upper threshold of mask
        blInvert: Color invert Yes or No (originally invert for blue mask)
        Show: working wndows Showing Mode
        TrFac: Transparency Factor
        Gaussian: Apply or not Gaussian Blur
        GaussianSigma: Gaussian Blur Sigma
        MaskInvert: Mask Inversion for excluding selected area (True / False)
        DoAdjLevel: adjust level between channels
        Rel2B: compare to B only mode option (True/False)
    """

#-------------------------------------------
class DialogPreviewer(ItemListener):
    def __init__(self, choice1, choice2):
        """
            choice: a java.awt.choice
        """
        self.choice1 = choice1
        self.choice2 = choice2


    def itemStateChanged(self, event):
        """ event: an ItemEvent with data on what happened to the checkbox. """
        choice_item  = self.choice1.getSelectedItem()
        #print "Dialog Previewr: ", str(choice_item)
        if choice_item == "Luminosity":
            choice_item2 = "Luminosity"
        elif choice_item == "Yellow/Blue":
            choice_item2 = "Blue"
        elif choice_item == "Cyan/Red":
            choice_item2 = "Red"
        else: # Magenta/Green
            choice_item2 = "Green"
        self.choice2.select(choice_item2)
        return

#-----------------------------
def selectChannel():
    # dialog of selecting target channel for mask valid lightness range
    gd = GenericDialog("Relative RGB Color Channel Mask Maker Research Ver."\
                                            + "/ Target Channel Selection")
    scolorCh = ["Luminosity", "Cyan/Red","Magenta/Green","Yellow/Blue"]
    gd.addChoice("Select target channel for making mask", \
                                                scolorCh, scolorCh[2])
    sbaseCh = ["Luminosity","Red","Green","Blue"]
    gd.addChoice("Select target channel for deciding mask valid " + \
                                    "lightness range", sbaseCh, sbaseCh[2])
    gd.addMessage("If you select Luminosity as making mask ch., \n"\
            + "  lightness range deciding ch. will be coerced to Luminosity")

    # The UI elements for the above two inputs
    choice1 = gd.getChoices().get(0)
    choice2 = gd.getChoices().get(1)
    #---- set Previewer ----
    previewer0 = DialogPreviewer(choice1, choice2)
    choice1.addItemListener(previewer0)
    #---- show the dialog ----
    gd.showDialog()

    colorCh =  gd.getNextChoice()
    baseCh =  gd.getNextChoice()
    if gd.wasCanceled():
        colorCh =  "None"
        baseCh =  "None"

    if colorCh == "Luminosity" and baseCh != "Luminosity":
        baseCh = "Luminosity"
        IJ.showMessage("lightness range deciding ch. " +\
                                    "is changed to Luminosity")
    #----remove Previewer-----
    choice1.removeItemListener(previewer0)

    return str(baseCh),  str(colorCh)
    """
    return values
        baseCh: channel name for deciding mask upper & lower thresholds
        color Ch: channel name for making color mask image
    """

#-------------------------------------------
def parameter_input(stackSize, sequence):
## opening parameter input dialog 1
#   Mode: Global Variable
#   parameter input for deciding base channel and color mask channel
    sequence = sequence + 1
    if stackSize == 1: # if imp0 is Monochrome
        baseCh = "Luminosity"
        colorCh = "Luminosity"
    else: # if imp0 is RGB Color
        baseCh, colorCh = selectChannel()
            # select target channel for deciding upper & lower threshold
        if baseCh == "None":
            imp0.hide()
            imp0.close()
    if colorCh == "Yellow/Blue":
        C0 = "Yellow"
        C1 = "Blue"
    elif colorCh == "Cyan/Red":
        C0 = "Cyan"
        C1 = "Red"
    elif colorCh == "Magenta/Green":
        C0 = "Magenta"
        C1 = "Green"
    elif colorCh == "Luminosity":
        C0 = "Luminosity Positive"
        C1 = "Negative"
    if baseCh != "None":
        imp0.hide()
#   --load preview image--
        if baseCh == "Blue":
            impPrvBase = Duplicator().run(impprvsCh[2])
        elif baseCh == "Red":
            impPrvBase = Duplicator().run(impprvsCh[0])
        elif baseCh == "Green":
            impPrvBase = Duplicator().run(impprvsCh[1])
        else:  # Luminosity
            impPrvBase = impPrvLuminosity # input Luminosity Gray8 image
        impPrvBase.setTitle("impPrvBase")
        impStat = impPrvBase.getStatistics()
        printStat(impStat, "impPrvBase")
        if C0 == "Luminosity Positive":
            parm = thresholdImageUIb\
                (impPrvBase, Mode, C0, C1, sequence, stackSize)
        else: # RGB mask
            parm = thresholdImageUIa(impPrvBase, impprvsCh[0], impprvsCh[1], \
                                        impprvsCh[2], Mode, C0, C1, sequence)

        impPrvBase.close()
## opening parameter input dialog 2
#    threshold parameters input for making masks
#          (for valid ligheness range or color threshold)
    else: # if parameter input dialog1 was canceled
        parm = []
        parm.append("Stop")
    return parm, baseCh, colorCh, stackSize, sequence

#----load image-------
def load_image():
    imgCurrent = WM.getCurrentImage()
    if not imgCurrent:
        od = OpenDialog("Choose a file")
        folder = od.getDirectory()
        filename = od.getFileName()
    # Always Open file with Bio-format Plugin
        if filename:
            imps = BF.openImagePlus(folder + filename)
            imp0 = imps[0] #imp0: Original Image
            imp0.show()
        else:
            IJ.showMessage("File is not selected.")
    else: # target image is already open
        imp0 = imgCurrent
        filename = imp0.getTitle()
        try:
            folder = imp0.getOriginalFileInfo().directory
        except:
            folder = IJ.getDirectory("Output_directory")
    return imp0, filename, folder
#------------ main routine -----------#
window = WM.getActiveWindow()
imp0, filename, folder = load_image()
Mode = "Auto"
# deciding bit depth
if Mode == "Auto":
    bitdepth = imp0.getBitDepth()
    if bitdepth == 8:
            Mode = "8-bit"
            Lmax = 255
    elif bitdepth ==16:
            Mode = "16-bit"
            Lmax = (256 ** 2) - 1
    elif bitdepth ==32:
            Mode = "32-bit"
            Lmax = set_Lmax(imp0)
            TmpStat = imp0.getStatistics()
            imp0.setDisplayRange(0.0, Lmax)
            printStat(TmpStat, "imp0 ")
    else: # 24 bit ColorProcessor
            Mode = "8-bit"
            Lmax = 255
stackSize =imp0.getStackSize() # Color or Monochrome?
if bitdepth == 24: # in case of colorProcessor
    stackSize = 3
if Mode != "Stop": # if user did not cancel executing process
    """ making preview image """
    impprv = Duplicator().run(imp0) # impprv: preview imp
    impprv.setDisplayRange(0.0, Lmax)
    IJ.run(impprv, "8-bit", "") # Preview Image converted to 8-bit
    scale = 1200.0 / impprv.getWidth()
                    # preview image scaling value (width = 1200px fixed)
    new_height = impprv.getHeight() * scale # preview image height
    new_height = int(new_height)
    tmpimpprv = impprv.resize(1200, new_height, "none")
    if stackSize == 3:
        impPrvLuminosity = convertToLuminosity(tmpimpprv)
    else: # mono channel
        impPrvLuminosity = Duplicator().run(tmpimpprv)
                        # converted to mono channel grayscale
    ImageConverter(impPrvLuminosity).convertToGray8() # converted to Gray8
    impprvsCh = ChannelSplitter.split(tmpimpprv)
    for i in impprvsCh:
        ImageConverter(i).convertToGray8() 
                                # making each channel 8bit preview images
        i.getProcessor().setLut(None)


    """ start for defining parameters and making masks process """
    sequence = 0  # 1st mask image
    parm, baseCh, colorCh, stackSize, sequence = \
                                    parameter_input(stackSize, sequence)
    if parm[0] != "Stop": # if parameter input dialog2 was not canceled
        """ proceeding 1st mask making process """
        TmpStat = imp0.getStatistics()
        printStat(TmpStat, "imp0 (2nd)")
        if stackSize == 3:
            impR, impG, impB, impLuminosity = \
                    RGBmanipulation(imp0, parm, stackSize)
            baseProsess(impR, impG, impB, impLuminosity, folder,\
                filename, baseCh, parm, colorCh, stackSize, sequence)
        else:
            impLuminosity = Duplicator().run(imp0)
            baseProsess(impLuminosity, impLuminosity, impLuminosity, \
                impLuminosity, folder, filename, baseCh, parm, colorCh, \
                                                    stackSize, sequence)

        if parm[10] == True and stackSize == 3:  # Making 2nd mask image
            parm, baseCh, colorCh, stackSize, sequence = \
                            parameter_input(stackSize, sequence)
            if parm[0] != "Stop": 
                # if parameter input dialog2 was not canceled
                """ proceeding 2nd mask making process """
                imps100 = BF.openImagePlus(folder + filename + \
                                                        "_tmpMask_1.tif")
                """
                impR, impG, impB, impLuminosity = \
                                    RGBmanipulation(imp0, parm, stackSize)
                """
                baseProsess(impR, impG, impB, impLuminosity, folder,\
                        filename, baseCh, parm, colorCh, stackSize, sequence)
                """ combining 2 mask images """
                if sequence ==2: # if there are 2 masks then combine 2 masks
                    imps101 = BF.openImagePlus(folder + filename + \
                                                    "_tmpMask_2.tif")
                    ic = plugin.ImageCalculator()
                    maskLastBindimp = ic.run("Min created", \
                                        imps100[0], imps101[0])  # bind masks
                    IJ.saveAsTiff(maskLastBindimp, folder + filename + \
                                                        "_combined_mask.tif")
#
    # closing temporary windows for preview images
    imp0.close()
    tmpimpprv.hide()
    tmpimpprv.close()
    impprvsCh[0].hide()
    impprvsCh[0].close()
    if stackSize == 3:
        impprvsCh[1].hide()
        impprvsCh[2].hide()
        impprvsCh[1].close()
        impprvsCh[2].close()
        impR.close()
        impG.close()
        impB.close()
    impLuminosity.close()

else: # if user canceled file opening
    IJ.log("User canceled")
IJ.log("Finish")

